#include "TamaP.h"

#include <math.h>

#include "etc.h"

#ifndef DEFAULT_PROBABILITY
#define DEFAULT_PROBABILITY 0.1
#endif

#ifndef DEFAULT_SIZE
#define DEFAULT_SIZE 3
#endif

#ifndef DEFAULT_AIR
#define DEFAULT_AIR 0.05
#endif

#ifndef DEFAULT_GRAVITY
#define DEFAULT_GRAVITY 0.3
#endif

#ifndef DEFAULT_TRANSMISSION
#define DEFAULT_TRANSMISSION 60.0
#endif

#ifndef DEFAULT_AFTER_IMAGE
#define DEFAULT_AFTER_IMAGE 10
#endif

/*===========================================================================*/
/* ֥ȤΥФμ                                                */
/*===========================================================================*/

/*---------------------------------------------------------------------------*/
/* TamaClass 饹                                                          */
/*---------------------------------------------------------------------------*/

char * TamaClass_GetName(TamaClass tama_class)
{ return (tama_class->name); }

double TamaClass_GetProbability(TamaClass tama_class)
{ return (tama_class->probability); }

ObjList TamaClass_GetGCListList(TamaClass tama_class)
{ return (tama_class->gc_list_list); }

ObjList TamaClass_GetNextList(TamaClass tama_class)
{ return (tama_class->next_list); }

/*---------------------------------------------------------------------------*/
/* TamaNext 饹                                                           */
/*---------------------------------------------------------------------------*/

TamaClass TamaNext_GetTamaClass(TamaNext tama_next)
{ return (tama_next->tama_class); }

double TamaNext_GetPower(TamaNext tama_next) { return (tama_next->power); }
int TamaNext_GetNumber(TamaNext tama_next) { return (tama_next->number); }

/*---------------------------------------------------------------------------*/
/* Tamas 饹                                                              */
/*---------------------------------------------------------------------------*/

TamaClass Tamas_GetTamaClass(Tamas tamas) { return (tamas->tama_class); }
int Tamas_GetNumber(Tamas tamas) { return (tamas->number); }
double * Tamas_GetX(Tamas tamas) { return (tamas->x); }
double * Tamas_GetY(Tamas tamas) { return (tamas->y); }
double * Tamas_GetZ(Tamas tamas) { return (tamas->z); }
double * Tamas_GetVx(Tamas tamas) { return (tamas->vx); }
double * Tamas_GetVy(Tamas tamas) { return (tamas->vy); }
double * Tamas_GetVz(Tamas tamas) { return (tamas->vz); }

/*===========================================================================*/
/* TamaClass ֥Ȥκ                                      */
/*===========================================================================*/

TamaClass TamaClass_Create(char * name,
			   int size,
			   double probability,
			   double air,
			   double gravity,
			   double transmission,
			   int after_image_length,
			   int fine,
			   ObjList gc_list_list,
			   ObjList next_list,
			   Disp disp)
{
  TamaClass tama_class;
  int i;

  tama_class = (TamaClass)malloc(sizeof(_TamaClass));
  if (!tama_class) Error("TamaClass_Create", "Cannot allocate memory.");

  if ((name == NULL) || (name[0] == '\0')) name = "Unknown\0";

  tama_class->name = (char *)malloc(sizeof(char) * (StringLen(name) + 1));
  if (tama_class->name == NULL)
    Error("TamaClass_Create", "Cannot allocate memory.");
  StringCpy(tama_class->name, name);

  tama_class->size               = size;
  tama_class->probability        = probability;
  tama_class->air                = air;
  tama_class->gravity            = gravity;
  tama_class->transmission       = transmission;
  tama_class->after_image_length = after_image_length;
  tama_class->fine               = fine;

  if (gc_list_list) tama_class->gc_list_list = gc_list_list;
  else              tama_class->gc_list_list = ObjList_Create();

  if (next_list) tama_class->next_list = next_list;
  else           tama_class->next_list = ObjList_Create();

  return (tama_class);
}

TamaClass TamaClass_Destroy(TamaClass tama_class)
{
  if (!tama_class) return (NULL);

  if (tama_class->name) free(tama_class->name);

  if (tama_class->gc_list_list)
    ObjList_Destroy(tama_class->gc_list_list);

  if (tama_class->next_list)
    ObjList_Destroy(tama_class->next_list);

  free(tama_class);

  return (NULL);
}

/*===========================================================================*/
/* TamaNext ֥Ȥκ                                       */
/*===========================================================================*/

TamaNext TamaNext_Create(TamaClass tama_class, double power, int n)
{
  TamaNext next;

  next = (TamaNext)malloc(sizeof(_TamaNext));
  if (!next) Error("TamaNext_Create", "Cannot allocate memory.");

  next->tama_class = tama_class;
  next->power = power;
  next->number = n;

  return (next);
}

TamaNext TamaNext_Destroy(TamaNext next)
{
  if (!next) return (NULL);

  free(next);

  return (NULL);
}

/*===========================================================================*/
/* ǻѤؿ                                                        */
/*===========================================================================*/

/*---------------------------------------------------------------------------*/
/* TamaClass θ                                                          */
/* ĤʤäˤϡTamaClass ֥Ȥ                */
/* ꥹȤɲä롥                                                        */
/*---------------------------------------------------------------------------*/

static TamaClass TamaClassList_SearchTamaClass(ObjList list,
					       char * name,
					       Disp disp,
					       int default_size,
					       double default_air,
					       double default_gravity,
					       double default_transmission,
					       int default_after_image,
					       int fine)
{
  ObjListData current;
  TamaClass tama_class;

  if (!list) Error("TamaClassList_SearchTamaClass",
		   "TamaClassList has not created.");

  for (current = ObjList_GetStart(list);
       !ObjList_IsEndEdge(list, current);
       current = ObjListData_GetNext(current)) {
    tama_class = (TamaClass)ObjListData_GetObj(current);
    if (!StringCmp(tama_class->name, name)) {
      return (tama_class);
    }
  }

  /* ĤʤäˤϡꥹȤɲä */
  tama_class = TamaClass_Create(name,
				default_size,
				-1.0,
				default_air,
				default_gravity,
				default_transmission,
				default_after_image,
				fine,
				NULL, NULL, disp);
  ObjList_InsertObjToEnd(list, tama_class, (ObjDestructor)TamaClass_Destroy);

  return (tama_class);
}

/*---------------------------------------------------------------------------*/
/* ñμ                                                                */
/*---------------------------------------------------------------------------*/

static char * GetWord(Stream stream)
{
  char * word;
  word = Stream_GetWord(stream, NULL, 0,
			"",     /* ʸڤʬ */
			" \t",  /* ʸǤڤʬ */
			"!#$%", /* ȹԤƬʸ '#' ʤ */
			"\n",   /* ʸ '\n' ʤ */
			"\"\'", /* ʸ '\"' ʤ */
			"",     /* ʸƬɤߤȤФʸ */
			""      /* ʸɤߤȤФʸ */
			);
  return (word);
}

static char * GetWord2(Stream stream)
{
  char * word;
  word = Stream_GetWord(stream, NULL, 0,
			"",     /* ʸڤʬ */
			" \t",  /* ʸǤڤʬ */
			"",     /* ȹԤƬʸ '#' ʤ */
			"\n",   /* ʸ '\n' ʤ */
			"\"\'", /* ʸ '\"' ʤ */
			"",     /* ʸƬɤߤȤФʸ */
			""      /* ʸɤߤȤФʸ */
			);
  return (word);
}

static ObjList ReadNext(ObjList tama_class_list, char * word, Disp disp,
			int default_size,
			double default_air,
			double default_gravity,
			double default_transmission,
			int default_after_image,
			int fine)
{
  ObjList list;
  Stream stream;
  char * w1;
  char * w2;
  char * w3;
  TamaClass tama_class;
  double power;
  int n;

  list = ObjList_Create();
  if (!word) return (list);
  stream = Stream_CreateFromCharacters(word);
  if (!stream) return (list);

  while ((w1 = GetWord2(stream)) != NULL) {
    w2 = w3 = NULL;
    tama_class = TamaClassList_SearchTamaClass(tama_class_list, w1, disp,
					       default_size,
					       default_air,
					       default_gravity,
					       default_transmission,
					       default_after_image,
					       fine);
    w2 = GetWord2(stream);
    if ((w3 = GetWord2(stream)) != NULL) {
      power = atof(w2);
      n = atoi(w3);
      ObjList_InsertObjToEnd(list, TamaNext_Create(tama_class, power, n),
			     (ObjDestructor)TamaNext_Destroy);
    }
    if (w1) free(w1);
    if (w2) free(w2);
    if (w3) free(w3);
  }

  Stream_Destroy(stream);

  return (list);
}

/*===========================================================================*/
/* ǡɤ߹                                                          */
/*===========================================================================*/

ObjList TamaClassList_CreateFromStream(Disp disp,
				       ColorGCDatabase database,
				       Stream stream,
				       char * background_color,
				       int fine)
{
  ObjList tama_class_list;
  TamaClass current = NULL;
  char * word;
  char * word2;
  ObjList color_gc_list;
  double default_probability  = DEFAULT_PROBABILITY;
  int default_size            = DEFAULT_SIZE;
  double default_air          = DEFAULT_AIR;
  double default_gravity      = DEFAULT_GRAVITY;
  double default_transmission = DEFAULT_TRANSMISSION;
  int default_after_image     = DEFAULT_AFTER_IMAGE;

  tama_class_list = ObjList_Create();

  while ((word = GetWord(stream)) != NULL) {
    word2 = NULL;

    if (!StringCmp(word, "Name")) {
      if ((word2 = GetWord(stream)) != NULL) {
	current =
	  TamaClassList_SearchTamaClass(tama_class_list, word2, disp,
					default_size, default_air,
					default_gravity,
					default_transmission,
					default_after_image,
					fine);
      }
    } else if (!StringCmp(word, "DefaultProbability")) {
      if ((word2 = GetWord(stream)) != NULL) {
	default_probability = atof(word2);
      }
    } else if (!StringCmp(word, "DefaultSize")) {
      if ((word2 = GetWord(stream)) != NULL) {
	default_size = atoi(word2);
      }
    } else if (!StringCmp(word, "DefaultAir")) {
      if ((word2 = GetWord(stream)) != NULL) {
	default_air = atof(word2);
      }
    } else if (!StringCmp(word, "DefaultGravity")) {
      if ((word2 = GetWord(stream)) != NULL) {
	default_gravity = atof(word2);
      }
    } else if (!StringCmp(word, "DefaultTransmission")) {
      if ((word2 = GetWord(stream)) != NULL) {
	default_transmission = atof(word2);
      }
    } else if (!StringCmp(word, "DefaultAfterImage")) {
      if ((word2 = GetWord(stream)) != NULL) {
	default_after_image = atoi(word2);
      }
    } else {
      if (current) {
	if (!StringCmp(word, "Size")) {
	  if ((word2 = GetWord(stream)) != NULL) {
	    if (!StringCmp(word2, "Default") || !StringCmp(word2, "default"))
	      current->size = default_size;
	    else current->size = atoi(word2);
	  }
	} else if (!StringCmp(word, "Probability")) {
	  if ((word2 = GetWord(stream)) != NULL) {
	    if (!StringCmp(word2, "Default") || !StringCmp(word2, "default"))
	      current->probability = default_probability;
	    else if (!StringCmp(word2, "None") || !StringCmp(word2, "none"))
	      current->probability = -1.0;
	    else current->probability = atof(word2);
	  }
	} else if (!StringCmp(word, "Air")) {
	  if ((word2 = GetWord(stream)) != NULL) {
	    if (!StringCmp(word2, "Default") || !StringCmp(word2, "default"))
	      current->air = default_air;
	    else current->air = atof(word2);
	  }
	} else if (!StringCmp(word, "Gravity")) {
	  if ((word2 = GetWord(stream)) != NULL) {
	    if (!StringCmp(word2, "Default") || !StringCmp(word2, "default"))
	      current->gravity = default_gravity;
	    else current->gravity = atof(word2);
	  }
	} else if (!StringCmp(word, "Transmission")) {
	  if ((word2 = GetWord(stream)) != NULL) {
	    if (!StringCmp(word2, "Default") || !StringCmp(word2, "default"))
	      current->transmission = default_transmission;
	    else current->transmission = atof(word2);
	  }
	} else if (!StringCmp(word, "AfterImage")) {
	  if ((word2 = GetWord(stream)) != NULL) {
	    if (!StringCmp(word2, "Default") || !StringCmp(word2, "default"))
	      current->after_image_length = default_after_image;
	    else current->after_image_length = atoi(word2);
	  }
	} else if (!StringCmp(word, "Color")) {
	  if ((word2 = GetWord(stream)) != NULL) {
	    if (current->gc_list_list == NULL)
	      current->gc_list_list = ObjList_Create();
	    if (!StringEqual(word2, "None") && !StringEqual(word2, "none")) {
	      color_gc_list =
		CreateColorGCListFromCharacters(database,
						word2,
						background_color,
						current->after_image_length,
						fine);
	      ObjList_Concatenate(current->gc_list_list, color_gc_list);
	    }
	  }
	} else if (!StringCmp(word, "Next")) {
	  if ((word2 = GetWord(stream)) != NULL) {
	    if (current->next_list == NULL)
	      current->next_list = ObjList_Create();
	    if (!StringEqual(word2, "None") && !StringEqual(word2, "none")) {
	      ObjList_Concatenate(current->next_list,
				  ReadNext(tama_class_list, word2, disp,
					   default_size,
					   default_air,
					   default_gravity,
					   default_transmission,
					   default_after_image,
					   fine));
	    }
	  }
	} else if (!StringCmp(word, "End")) {
	  if (word)  free(word);
	  if (word2) free(word2);
	  return (tama_class_list);
	} else {
	  fprintf(stderr, "Unknown command : %s\n", word);
	}
      } else {
	fprintf(stderr,
		"Cannot specify class name. Command was ignored. : %s\n",
		word);
      }
    }
    if (word)  free(word);
    if (word2) free(word2);
  }

  return (tama_class_list);
}

/*===========================================================================*/
/* Tamas ֥Ȥ                                                  */
/*===========================================================================*/

/*---------------------------------------------------------------------------*/
/* Tamas ֥Ȥ                                                  */
/*---------------------------------------------------------------------------*/

Tamas Tamas_Create(TamaClass tama_class, double power, int number, int n,
		   double * x, double * y, double * z,
		   double * vx, double * vy, double * vz,
		   int x_min, int y_min, int x_max, int y_max)
{
  double rad, z_rad;
  Tamas tamas;
  double p, t;
  int i, j, a;

  tamas = (Tamas)malloc(sizeof(_Tamas));
  if (!tamas) Error("Tamas_Create", "Cannot allocate memory");

  tamas->tama_class = tama_class;

  tamas->x  = (double *)malloc(sizeof(double) * number * n);
  if (tamas->x  == NULL) Error("Tamas_Create", "Cannot allocate memory");

  tamas->y  = (double *)malloc(sizeof(double) * number * n);
  if (tamas->y  == NULL) Error("Tamas_Create", "Cannot allocate memory");

  tamas->z  = (double *)malloc(sizeof(double) * number * n);
  if (tamas->z  == NULL) Error("Tamas_Create", "Cannot allocate memory");

  tamas->vx = (double *)malloc(sizeof(double) * number * n);
  if (tamas->vx == NULL) Error("Tamas_Create", "Cannot allocate memory");

  tamas->vy = (double *)malloc(sizeof(double) * number * n);
  if (tamas->vy == NULL) Error("Tamas_Create", "Cannot allocate memory");

  tamas->vz = (double *)malloc(sizeof(double) * number * n);
  if (tamas->vz == NULL) Error("Tamas_Create", "Cannot allocate memory");

  tamas->gc_list = ObjList_GetStart(tamas->tama_class->gc_list_list);

  a = 0;
  for (i = 0; i < n; i++) {
    for (j = 0; j < number; j++) {
      z_rad = DoubleRand(M_PI * 2.0);
      p = cos(z_rad);
      rad = DoubleRand(M_PI * 2.0);
      tamas->x[a] = x[i];
      tamas->y[a] = y[i];
      tamas->z[a] = z[i];
      t = tamas->tama_class->transmission * 0.01;
      tamas->vx[a] = cos(rad) * power * p + vx[i] * t;
      tamas->vy[a] = sin(rad) * power * p + vy[i] * t;
      tamas->vz[a] = power * sin(z_rad)   + vz[i] * t;
      if ( ((tamas->x[a] < x_min) && (tamas->vx[a] < 0.0)) ||
	   ((tamas->x[a] > x_max) && (tamas->vx[a] > 0.0)) ||
	   ((tamas->y[a] > y_max) && (tamas->vy[a] > 0.0)) ) {
	/* None */
      } else {
	a++;
      }
    }
  }
  tamas->number = a;

  return (tamas);
}

/*---------------------------------------------------------------------------*/
/* Tama ֥Ȥκ                                                   */
/*---------------------------------------------------------------------------*/

Tamas Tamas_Destroy(Tamas tamas)
{
  if (!tamas) return (NULL);

  if (tamas->x ) free(tamas->x );
  if (tamas->y ) free(tamas->y );
  if (tamas->z ) free(tamas->z );
  if (tamas->vx) free(tamas->vx);
  if (tamas->vy) free(tamas->vy);
  if (tamas->vz) free(tamas->vz);

  free(tamas);

  return (NULL);
}

/*---------------------------------------------------------------------------*/
/* Tamas ֥Ȥΰư                                                  */
/*---------------------------------------------------------------------------*/

AfterImages Tamas_Move(Tamas tamas,
		       int x_min, int y_min, int x_max, int y_max, int size)
{
  AfterImages after_images;
  int length;
  int i, a;
  double z;
  double step;

  if (ObjList_IsEndEdge(tamas->tama_class->gc_list_list, tamas->gc_list))
    return (NULL);

  step = 100.0 / tamas->tama_class->fine;

  a = 0;
  for (i = 0; i < tamas->number; i++) {
    if (a) {
      tamas->x[ i] = tamas->x[ i + a];
      tamas->y[ i] = tamas->y[ i + a];
      tamas->z[ i] = tamas->z[ i + a];
      tamas->vx[i] = tamas->vx[i + a];
      tamas->vy[i] = tamas->vy[i + a];
      tamas->vz[i] = tamas->vz[i + a];
    }

    tamas->vy[i] += tamas->tama_class->gravity * step;

    /* 񹳤η׻ */
    tamas->vx[i] *= (1.0 - tamas->tama_class->air * step);
    tamas->vy[i] *= (1.0 - tamas->tama_class->air * step);
    tamas->vz[i] *= (1.0 - tamas->tama_class->air * step);

    /* Ԥη׻ */
    tamas->z[i] += tamas->vz[i] * step;
    z = 1.0 - tamas->z[i] * (1.0 / 1000.0);
    if (z < 0.01) z = 0.01;

    tamas->x[i] += tamas->vx[i] * step * z * size * 0.001;
    tamas->y[i] += tamas->vy[i] * step * z * size * 0.001;

    if (((tamas->x[i] < x_min) && (tamas->vx[i] < 0.0)) ||
	((tamas->x[i] > x_max) && (tamas->vx[i] > 0.0)) ||
	(tamas->y[i] > y_max)) {
      i--;
      a++;
      tamas->number--;
    }
  }

  after_images = AfterImages_Create(tamas->tama_class->size, tamas->number,
				    tamas->x, tamas->y,
				    x_min, y_min, x_max, y_max,
				    ObjListData_GetObj(tamas->gc_list));

  tamas->gc_list = ObjListData_GetNext(tamas->gc_list);

  return (after_images);
}

/*****************************************************************************/
/* End of Program                                                            */
/*****************************************************************************/
